import functools
import inspect
import os
import time
from typing import Any

from smolagents import CodeAgent, LiteLLMModel, tool

from palimpzest.core.data.context import Context
from palimpzest.core.data.context_manager import ContextManager
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.models import GenerationStats, OperatorCostEstimates, RecordOpStats
from palimpzest.query.operators.physical import PhysicalOperator

# TODO: need to store final executed code in compute() operator so that humans can debug when human-in-the-loop

def make_tool(bound_method):
    # Get the original function and bound instance
    func = bound_method.__func__
    instance = bound_method.__self__
    
    # Get the signature and remove 'self'
    sig = inspect.signature(func)
    params = list(sig.parameters.values())[1:]  # skip 'self'
    new_sig = inspect.Signature(parameters=params, return_annotation=sig.return_annotation)

    # Create a wrapper function dynamically
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(instance, *args, **kwargs)

    # Update the __signature__ to reflect the new one without 'self'
    wrapper.__signature__ = new_sig

    return wrapper


class SmolAgentsCompute(PhysicalOperator):
    """
    """
    def __init__(self, context_id: str, instruction: str, additional_contexts: list[Context] | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context_id = context_id
        self.instruction = instruction
        self.additional_contexts = [] if additional_contexts is None else additional_contexts
        # self.model_id = "anthropic/claude-3-7-sonnet-latest"
        self.model_id = "openai/gpt-4o-mini-2024-07-18"
        # self.model_id = "openai/gpt-4o-2024-08-06"
        api_key = os.getenv("ANTHROPIC_API_KEY") if "anthropic" in self.model_id else os.getenv("OPENAI_API_KEY")
        self.model = LiteLLMModel(model_id=self.model_id, api_key=api_key)

    def __str__(self):
        op = super().__str__()
        op += f"    Context ID: {self.context_id:20s}\n"
        op += f"    Instruction: {self.instruction:20s}\n"
        op += f"    Add. Ctxs: {self.additional_contexts}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {
            "context_id": self.context_id,
            "instruction": self.instruction,
            "additional_contexts": self.additional_contexts,
            **id_params,
        }

    def get_op_params(self):
        op_params = super().get_op_params()
        return {
            "context_id": self.context_id,
            "instruction": self.instruction,
            "additional_contexts": self.additional_contexts,
            **op_params,
        }

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        return OperatorCostEstimates(
            cardinality=source_op_cost_estimates.cardinality,
            time_per_record=100,
            cost_per_record=1,
            quality=1.0,
        )

    def _create_record_set(
        self,
        candidate: DataRecord,
        generation_stats: GenerationStats,
        total_time: float,
        answer: dict[str, Any],
    ) -> DataRecordSet:
        """
        Given an input DataRecord and a determination of whether it passed the filter or not,
        construct the resulting RecordSet.
        """
        # create new DataRecord
        data_item = {field: answer[field] for field in self.output_schema.model_fields if field in answer}
        dr = DataRecord.from_parent(self.output_schema, data_item, parent_record=candidate)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=total_time,
            cost_per_record=generation_stats.cost_per_record,
            model_name=self.get_model_name(),
            total_input_tokens=generation_stats.total_input_tokens,
            total_output_tokens=generation_stats.total_output_tokens,
            total_input_cost=generation_stats.total_input_cost,
            total_output_cost=generation_stats.total_output_cost,
            llm_call_duration_secs=generation_stats.llm_call_duration_secs,
            fn_call_duration_secs=generation_stats.fn_call_duration_secs,
            total_llm_calls=generation_stats.total_llm_calls,
            total_embedding_llm_calls=generation_stats.total_embedding_llm_calls,
            answer={k: v.description if isinstance(v, Context) else v for k, v in answer.items()},
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])

    def __call__(self, candidate: DataRecord) -> Any:
        start_time = time.time()

        # get the input context object and its tools
        input_context: Context = candidate.context
        description = input_context.description
        tools = [tool(make_tool(f)) for f in input_context.tools]

        # update the description to include any additional contexts
        for ctx in self.additional_contexts:
            # TODO: remove additional context if it is an ancestor of the input context
            # (not just if it is equal to the input context)
            if ctx.id == input_context.id:
                continue
            description += f"\n\nHere is some additional Context which may be useful:\n\n{ctx.description}"

        # perform the computation
        instructions = f"\n\nHere is a description of the Context whose data you will be working with, as well as any previously computed results:\n\n{description}"
        agent = CodeAgent(
            tools=tools,
            model=self.model,
            add_base_tools=False,
            instructions=instructions,
            return_full_result=True,
            additional_authorized_imports=["pandas", "io", "os"],
            planning_interval=4,
            max_steps=30,
        )
        result = agent.run(self.instruction)
        # NOTE: you can see the system prompt with `agent.memory.system_prompt.system_prompt`
        # full_steps = agent.memory.get_full_steps()

        # compute generation stats
        response = result.output
        input_tokens = result.token_usage.input_tokens
        output_tokens = result.token_usage.output_tokens
        cost_per_input_token = (3.0 / 1e6) if "anthropic" in self.model_id else (0.15 / 1e6) # (2.5 / 1e6) #
        cost_per_output_token = (15.0 / 1e6) if "anthropic" in self.model_id else (0.6 / 1e6) # (10.0 / 1e6) #
        input_cost = input_tokens * cost_per_input_token
        output_cost = output_tokens * cost_per_output_token
        generation_stats = GenerationStats(
            model_name=self.model_id,
            total_input_tokens=input_tokens,
            total_output_tokens=output_tokens,
            total_input_cost=input_cost,
            total_output_cost=output_cost,
            cost_per_record=input_cost + output_cost,
            llm_call_duration_secs=time.time() - start_time,
        )

        # update the description of the computed Context to include the result
        new_description = f"RESULT: {response}\n\n"
        cm = ContextManager()
        cm.update_context(id=self.context_id, description=new_description)

        # create and return record set
        field_answers = {
            "context": cm.get_context(id=self.context_id),
            f"result-{self.context_id}": response,
        }
        record_set = self._create_record_set(
            candidate,
            generation_stats,
            time.time() - start_time,
            field_answers,
        )

        return record_set

# import json; json.dumps(agent.memory.get_full_steps())
# agent.memory.get_full_steps()[1].keys()
# dict_keys(['step_number', 'timing', 'model_input_messages', 'tool_calls', 'error', 'model_output_message', 'model_output', 'code_action', 'observations', 'observations_images', 
# 'action_output', 'token_usage', 'is_final_answer'])
# agent.memory.get_full_steps()[1]['action_output']


"""This file contains prompts for MixtureOfAgentsConvert operations."""

### SYSTEM PROMPTS ###
MAP_MOA_PROPOSER_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a set of output fields to generate. Your task is to generate a detailed and succinct analysis describing what you believe is the correct value for each output field.
Be sure to cite information from the context as evidence of why your answers are correct. Do not hallucinate evidence.

You will be provided with a description of each input field and each output field.

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

OUTPUT FIELDS:
{example_output_fields}

CONTEXT:
{{{example_context}}}{image_disclaimer}{audio_disclaimer}

Let's think step-by-step in order to answer the question.

ANSWER: {example_answer}
---
"""

FILTER_MOA_PROPOSER_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a filter condition. Your task is to generate a detailed and succinct analysis describing whether you believe the input satisfies the filter condition.
Be sure to cite information from the context as evidence of why your determination is correct. Do not hallucinate evidence.

You will be provided with a description of each input field.

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

CONTEXT:
{{{example_context}}}{image_disclaimer}{audio_disclaimer}

FILTER CONDITION: {example_filter_condition}

Let's think step-by-step in order to answer the question.

ANSWER: {example_answer}
---
"""

### USER / INSTANCE-SPECIFIC PROMPTS ###
MAP_MOA_PROPOSER_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a set of output fields to generate. Your task is to generate a detailed and succinct analysis describing what you believe is the correct value for each output field.
Be sure to cite information from the context as evidence of why your answers are correct. Do not hallucinate evidence.
{desc_section}
You will be provided with a description of each input field and each output field.
---
INPUT FIELDS:
{input_fields_desc}

OUTPUT FIELDS:
{output_fields_desc}

CONTEXT:
{context}<<image-audio-placeholder>>

Let's think step-by-step in order to answer the question.

ANSWER: """

FILTER_MOA_PROPOSER_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a filter condition. Your task is to generate a detailed and succinct analysis describing whether you believe the input satisfies the filter condition.
Be sure to cite information from the context as evidence of why your determination is correct. Do not hallucinate evidence.
{desc_section}
You will be provided with a description of each input field.

An example is shown below:
---
INPUT FIELDS:
{input_fields_desc}

CONTEXT:
{context}<<image-audio-placeholder>>

FILTER CONDITION: {filter_condition}

Let's think step-by-step in order to answer the question.

ANSWER: """



import json
import textwrap
import time
from collections.abc import Generator
from typing import TYPE_CHECKING, Any

from rich.console import Group
from rich.live import Live
from rich.markdown import Markdown
from rich.rule import Rule
from rich.text import Text

if TYPE_CHECKING:
    import PIL.Image

from smolagents.agent_types import handle_agent_output_types
from smolagents.agents import (
    ActionOutput,
    CodeAgent,
    FinalAnswerPromptTemplate,
    ManagedAgentPromptTemplate,
    PlanningPromptTemplate,
    PromptTemplates,
    RunResult,
    ToolOutput,
    populate_template,
)
from smolagents.local_python_executor import fix_final_answer_code
from smolagents.memory import (
    ActionStep,
    FinalAnswerStep,
    PlanningStep,
    SystemPromptStep,
    TaskStep,
    Timing,
    TokenUsage,
    ToolCall,
)
from smolagents.models import (
    CODEAGENT_RESPONSE_FORMAT,
    ChatMessage,
    ChatMessageStreamDelta,
    MessageRole,
    agglomerate_stream_deltas,
)
from smolagents.monitoring import YELLOW_HEX, LogLevel
from smolagents.utils import (
    AgentError,
    AgentExecutionError,
    AgentGenerationError,
    AgentMaxStepsError,
    AgentParsingError,
    extract_code_from_text,
    parse_code_blobs,
    truncate_content,
)

from palimpzest.prompts import (
    CODE_AGENT_SYSTEM_PROMPT,
    DATA_DISCOVERY_AGENT_INITIAL_PLAN_PROMPT,
    DATA_DISCOVERY_AGENT_REPORT_PROMPT,
    DATA_DISCOVERY_AGENT_TASK_PROMPT,
    DATA_DISCOVERY_AGENT_UPDATE_PLAN_POST_MESSAGES_PROMPT,
    DATA_DISCOVERY_AGENT_UPDATE_PLAN_PRE_MESSAGES_PROMPT,
    FINAL_ANSWER_POST_MESSAGES_PROMPT,
    FINAL_ANSWER_PRE_MESSAGES_PROMPT,
)


# TODO: make this use memory the way you want
class PZBaseAgent(CodeAgent):
    def __init__(self, run_id: str, context_description: str, *args, **kwargs):
        # memory_config = {
        #     "vector_store": {
        #         "provider": "chroma",
        #         "config": {
        #             "collection_name": f"palimpzest-memory-{self.__class__.__name__}",
        #             "path": "./pz-chroma",
        #         }
        #     }
        # }
        # self.pz_memory = Memory.from_config(memory_config)
        self.run_id = run_id
        self.context_description = context_description
        super().__init__(*args, **kwargs)

    def write_memory_to_messages(
        self,
        summary_mode: bool = False,
    ) -> list[ChatMessage]:
        """
        Reads past llm_outputs, actions, and observations or errors from the memory into a series of messages
        that can be used as input to the LLM. Adds a number of keywords (such as PLAN, error, etc) to help
        the LLM.
        """
        messages = self.memory.system_prompt.to_messages(summary_mode=summary_mode)
        for memory_step in self.memory.steps:
            messages.extend(memory_step.to_messages(summary_mode=summary_mode))
        return messages

    def _generate_planning_step(
        self, task, is_first_step: bool, step: int
    ) -> Generator[ChatMessageStreamDelta | PlanningStep]:
        start_time = time.time()
        if is_first_step:
            input_messages = [
                ChatMessage(
                    role=MessageRole.USER,
                    content=[
                        {
                            "type": "text",
                            "text": populate_template(
                                self.prompt_templates["planning"]["initial_plan"],
                                variables={"task": task, "tools": self.tools, "managed_agents": self.managed_agents, "context_description": self.context_description},
                            ),
                        }
                    ],
                )
            ]
            if self.stream_outputs and hasattr(self.model, "generate_stream"):
                plan_message_content = ""
                output_stream = self.model.generate_stream(input_messages, stop_sequences=["<end_plan>"])  # type: ignore
                input_tokens, output_tokens = 0, 0
                with Live("", console=self.logger.console, vertical_overflow="visible") as live:
                    for event in output_stream:
                        if event.content is not None:
                            plan_message_content += event.content
                            live.update(Markdown(plan_message_content))
                            if event.token_usage:
                                output_tokens += event.token_usage.output_tokens
                                input_tokens = event.token_usage.input_tokens
                        yield event
            else:
                plan_message = self.model.generate(input_messages, stop_sequences=["<end_plan>"])
                plan_message_content = plan_message.content
                input_tokens, output_tokens = (
                    (
                        plan_message.token_usage.input_tokens,
                        plan_message.token_usage.output_tokens,
                    )
                    if plan_message.token_usage
                    else (None, None)
                )
            plan = textwrap.dedent(
                f"""Here are the facts I know and the plan of action that I will follow to solve the task:\n```\n{plan_message_content}\n```"""
            )
        else:
            # Summary mode removes the system prompt and previous planning messages output by the model.
            # Removing previous planning messages avoids influencing too much the new plan.
            memory_messages = self.write_memory_to_messages(summary_mode=True)
            plan_update_pre = ChatMessage(
                role=MessageRole.SYSTEM,
                content=[
                    {
                        "type": "text",
                        "text": populate_template(
                            self.prompt_templates["planning"]["update_plan_pre_messages"], variables={"task": task, "context_description": self.context_description}
                        ),
                    }
                ],
            )
            plan_update_post = ChatMessage(
                role=MessageRole.USER,
                content=[
                    {
                        "type": "text",
                        "text": populate_template(
                            self.prompt_templates["planning"]["update_plan_post_messages"],
                            variables={
                                "task": task,
                                "tools": self.tools,
                                "managed_agents": self.managed_agents,
                                "remaining_steps": (self.max_steps - step),
                                "context_description": self.context_description,
                            },
                        ),
                    }
                ],
            )
            input_messages = [plan_update_pre] + memory_messages + [plan_update_post]
            if self.stream_outputs and hasattr(self.model, "generate_stream"):
                plan_message_content = ""
                input_tokens, output_tokens = 0, 0
                with Live("", console=self.logger.console, vertical_overflow="visible") as live:
                    for event in self.model.generate_stream(
                        input_messages,
                        stop_sequences=["<end_plan>"],
                    ):  # type: ignore
                        if event.content is not None:
                            plan_message_content += event.content
                            live.update(Markdown(plan_message_content))
                            if event.token_usage:
                                output_tokens += event.token_usage.output_tokens
                                input_tokens = event.token_usage.input_tokens
                        yield event
            else:
                plan_message = self.model.generate(input_messages, stop_sequences=["<end_plan>"])
                plan_message_content = plan_message.content
                if plan_message.token_usage is not None:
                    input_tokens, output_tokens = (
                        plan_message.token_usage.input_tokens,
                        plan_message.token_usage.output_tokens,
                    )
            plan = textwrap.dedent(
                f"""I still need to solve the task I was given:\n```\n{self.task}\n```\n\nHere are the facts I know and my new/updated plan of action to solve the task:\n```\n{plan_message_content}\n```"""
            )
        log_headline = "Initial plan" if is_first_step else "Updated plan"
        self.logger.log(Rule(f"[bold]{log_headline}", style="orange"), Text(plan), level=LogLevel.INFO)
        yield PlanningStep(
            model_input_messages=input_messages,
            plan=plan,
            model_output_message=ChatMessage(role=MessageRole.ASSISTANT, content=plan_message_content),
            token_usage=TokenUsage(input_tokens=input_tokens, output_tokens=output_tokens),
            timing=Timing(start_time=start_time, end_time=time.time()),
        )

    # def _curate_messages(self, input_messages: list[ChatMessage]) -> list[ChatMessage]:
    #     """
    #     Try returning:
    #     - System Prompt + task
    #     - Current Plan
    #     - Summary of previous conversation
    #     """
    #     # initialize with the system prompt & original task
    #     curated_messages = input_messages[:2]

    #     # find the last planning step message
    #     idx = len(self.memory.steps) - 1
    #     while idx > -1:
    #         step = self.memory.steps[idx]
    #         if isinstance(step, PlanningStep):
    #             curated_messages.append(step.model_output_message)
    #             break
    #         idx -= 1

    #     # add summary of chat history
    #     history = self.pz_memory.search("A condensed summary of the execution trace of the agent.", run_id=self.run_id)
    #     for msg in history["results"]:
    #         pass

    #     return curated_messages

    def _step_stream(self, memory_step: ActionStep) -> Generator[ChatMessageStreamDelta | ActionOutput | ToolOutput]:
        """
        Perform one step in the ReAct framework: the agent thinks, acts, and observes the result.
        Yields ChatMessageStreamDelta during the run if streaming is enabled.
        At the end, yields either None if the step is not final, or the final answer.
        """
        memory_messages = self.write_memory_to_messages()

        input_messages = memory_messages.copy()

        ### Generate model output ###
        memory_step.model_input_messages = input_messages
        try:
            additional_args: dict[str, Any] = {}
            if self.grammar:
                additional_args["grammar"] = self.grammar
            if self._use_structured_outputs_internally:
                additional_args["response_format"] = CODEAGENT_RESPONSE_FORMAT
            if self.stream_outputs:
                output_stream = self.model.generate_stream(
                    input_messages,
                    stop_sequences=["<end_code>", "Observation:", "Calling tools:"],
                    **additional_args,
                )
                chat_message_stream_deltas: list[ChatMessageStreamDelta] = []
                with Live("", console=self.logger.console, vertical_overflow="visible") as live:
                    for event in output_stream:
                        chat_message_stream_deltas.append(event)
                        live.update(
                            Markdown(agglomerate_stream_deltas(chat_message_stream_deltas).render_as_markdown())
                        )
                        yield event
                chat_message = agglomerate_stream_deltas(chat_message_stream_deltas)
                memory_step.model_output_message = chat_message
                output_text = chat_message.content
            else:
                chat_message: ChatMessage = self.model.generate(
                    input_messages,
                    stop_sequences=["<end_code>", "Observation:", "Calling tools:"],
                    **additional_args,
                )
                memory_step.model_output_message = chat_message
                output_text = chat_message.content
                self.logger.log_markdown(
                    content=output_text,
                    title="Output message of the LLM:",
                    level=LogLevel.DEBUG,
                )

            # This adds <end_code> sequence to the history.
            # This will nudge ulterior LLM calls to finish with <end_code>, thus efficiently stopping generation.
            if output_text and output_text.strip().endswith("```"):
                output_text += "<end_code>"
                memory_step.model_output_message.content = output_text

            memory_step.token_usage = chat_message.token_usage
            memory_step.model_output = output_text
        except Exception as e:
            raise AgentGenerationError(f"Error in generating model output:\n{e}", self.logger) from e

        ### Parse output ###
        try:
            if self._use_structured_outputs_internally:
                code_action = json.loads(output_text)["code"]
                code_action = extract_code_from_text(code_action) or code_action
            else:
                code_action = parse_code_blobs(output_text)
            code_action = fix_final_answer_code(code_action)
            memory_step.code_action = code_action
        except Exception as e:
            error_msg = f"Error in code parsing:\n{e}\nMake sure to provide correct code blobs."
            raise AgentParsingError(error_msg, self.logger) from e

        memory_step.tool_calls = [
            ToolCall(
                name="python_interpreter",
                arguments=code_action,
                id=f"call_{len(self.memory.steps)}",
            )
        ]

        ### Execute action ###
        self.logger.log_code(title="Executing parsed code:", content=code_action, level=LogLevel.INFO)
        is_final_answer = False
        try:
            output, execution_logs, is_final_answer = self.python_executor(code_action)
            execution_outputs_console = []
            if len(execution_logs) > 0:
                execution_outputs_console += [
                    Text("Execution logs:", style="bold"),
                    Text(execution_logs),
                ]
            observation = "Execution logs:\n" + execution_logs
        except Exception as e:
            if hasattr(self.python_executor, "state") and "_print_outputs" in self.python_executor.state:
                execution_logs = str(self.python_executor.state["_print_outputs"])
                if len(execution_logs) > 0:
                    execution_outputs_console = [
                        Text("Execution logs:", style="bold"),
                        Text(execution_logs),
                    ]
                    memory_step.observations = "Execution logs:\n" + execution_logs
                    self.logger.log(Group(*execution_outputs_console), level=LogLevel.INFO)
            error_msg = str(e)
            if "Import of " in error_msg and " is not allowed" in error_msg:
                self.logger.log(
                    "[bold red]Warning to user: Code execution failed due to an unauthorized import - Consider passing said import under `additional_authorized_imports` when initializing your CodeAgent.",
                    level=LogLevel.INFO,
                )
            raise AgentExecutionError(error_msg, self.logger) from e

        truncated_output = truncate_content(str(output))
        observation += "Last output from code snippet:\n" + truncated_output
        memory_step.observations = observation
        
        # # TODO: add output to self.pz_memory
        # def get_role(msg_role):
        #     return str(msg_role).split(".")[-1].lower()

        # messages = [
        #     {"role": get_role(memory_step.model_output_message.role), "content": memory_step.model_output_message.content},
        #     {"role": "user", "content": memory_step.observations},
        # ]
        # self.pz_memory.add(messages, run_id=self.run_id, agent_id=self.name)

        execution_outputs_console += [
            Text(
                f"{('Out - Final answer' if is_final_answer else 'Out')}: {truncated_output}",
                style=(f"bold {YELLOW_HEX}" if is_final_answer else ""),
            ),
        ]
        self.logger.log(Group(*execution_outputs_console), level=LogLevel.INFO)
        memory_step.action_output = output
        yield ActionOutput(output=output, is_final_answer=is_final_answer)

    def _run_stream(
        self, task: str, max_steps: int, images: list["PIL.Image.Image"] | None = None
    ) -> Generator[ActionStep | PlanningStep | FinalAnswerStep | ChatMessageStreamDelta]:
        """
        Execute the agent.
        """
        self.step_number = 1
        returned_final_answer = False
        while not returned_final_answer and self.step_number <= self.max_steps:
            # Run a planning step if scheduled
            if self.planning_interval is not None and (
                self.step_number == 1 or (self.step_number - 1) % self.planning_interval == 0
            ):
                planning_start_time = time.time()
                planning_step = None
                for element in self._generate_planning_step(
                    self.task, is_first_step=len(self.memory.steps) == 1, step=self.step_number
                ):  # Don't use the attribute step_number here, because there can be steps from previous runs
                    yield element
                    planning_step = element
                assert isinstance(planning_step, PlanningStep)  # Last yielded element should be a PlanningStep
                self.memory.steps.append(planning_step)
                planning_end_time = time.time()
                planning_step.timing = Timing(
                    start_time=planning_start_time,
                    end_time=planning_end_time,
                )

            # Start action step!
            action_step_start_time = time.time()
            action_step = ActionStep(
                step_number=self.step_number,
                timing=Timing(start_time=action_step_start_time),
                observations_images=images,
            )
            self.logger.log_rule(f"Step {self.step_number}", level=LogLevel.INFO)
            try:
                for output in self._step_stream(action_step):
                    # Yield streaming deltas
                    if not isinstance(output, (ActionOutput, ToolOutput)):
                        # non-action, non-tool output
                        yield output

                    if isinstance(output, (ActionOutput, ToolOutput)) and output.is_final_answer:
                        if self.final_answer_checks:
                            self._validate_final_answer(output.output)
                        returned_final_answer = True
                        action_step.is_final_answer = True
                        final_answer = output.output
                        # handle final step
            except AgentGenerationError as e:
                # Agent generation errors are not caused by a Model error but an implementation error: so we should raise them and exit.
                raise e
            except AgentError as e:
                # Other AgentError types are caused by the Model, so we should log them and iterate.
                action_step.error = e
            finally:
                self._finalize_step(action_step)
                self.memory.steps.append(action_step)
                yield action_step
                self.step_number += 1

        if not returned_final_answer and self.step_number == self.max_steps + 1:
            final_answer = self._handle_max_steps_reached(self.task, images)
            yield action_step
        yield FinalAnswerStep(handle_agent_output_types(final_answer))

    def run(
        self,
        task: str,
        stream: bool = False,
        reset: bool = True,
        images: list["PIL.Image.Image"] | None = None,
        additional_args: dict | None = None,
        max_steps: int | None = None,
    ):
        """
        Run the agent for the given task.

        Args:
            task (`str`): Task to perform.
            stream (`bool`): Whether to run in streaming mode.
                If `True`, returns a generator that yields each step as it is executed. You must iterate over this generator to process the individual steps (e.g., using a for loop or `next()`).
                If `False`, executes all steps internally and returns only the final answer after completion.
            reset (`bool`): Whether to reset the conversation or keep it going from previous run.
            images (`list[PIL.Image.Image]`, *optional*): Image(s) objects.
            additional_args (`dict`, *optional*): Any other variables that you want to pass to the agent run, for instance images or dataframes. Give them clear names!
            max_steps (`int`, *optional*): Maximum number of steps the agent can take to solve the task. if not provided, will use the agent's default value.

        Example:
        ```py
        from smolagents import CodeAgent
        agent = CodeAgent(tools=[])
        agent.run("What is the result of 2 power 3.7384?")
        ```
        """
        max_steps = max_steps or self.max_steps
        self.task = task
        self.interrupt_switch = False
        if additional_args is not None:
            self.state.update(additional_args)
            self.task += f"""
You have been provided with these additional arguments, that you can access using the keys as variables in your python code:
{str(additional_args)}."""

        self.memory.system_prompt = SystemPromptStep(system_prompt=self.system_prompt)
        if reset:
            self.memory.reset()
            self.monitor.reset()

        self.logger.log_task(
            content=self.task.strip(),
            subtitle=f"{type(self.model).__name__} - {(self.model.model_id if hasattr(self.model, 'model_id') else '')}",
            level=LogLevel.INFO,
            title=self.name if hasattr(self, "name") else None,
        )
        self.memory.steps.append(TaskStep(task=self.task, task_images=images))

        if getattr(self, "python_executor", None):
            self.python_executor.send_variables(variables=self.state)
            self.python_executor.send_tools({**self.tools, **self.managed_agents})

        if stream:
            # The steps are returned as they are executed through a generator to iterate on.
            return self._run_stream(task=self.task, max_steps=max_steps, images=images)
        run_start_time = time.time()
        # Outputs are returned only at the end. We only look at the last step.

        steps = list(self._run_stream(task=self.task, max_steps=max_steps, images=images))
        assert isinstance(steps[-1], FinalAnswerStep)
        output = steps[-1].output

        if self.return_full_result:
            total_input_tokens = 0
            total_output_tokens = 0
            correct_token_usage = True
            for step in self.memory.steps:
                if isinstance(step, (ActionStep, PlanningStep)):
                    if step.token_usage is None:
                        correct_token_usage = False
                        break
                    else:
                        total_input_tokens += step.token_usage.input_tokens
                        total_output_tokens += step.token_usage.output_tokens
            if correct_token_usage:
                token_usage = TokenUsage(input_tokens=total_input_tokens, output_tokens=total_output_tokens)
            else:
                token_usage = None

            if self.memory.steps and isinstance(getattr(self.memory.steps[-1], "error", None), AgentMaxStepsError):
                state = "max_steps_error"
            else:
                state = "success"

            messages = self.memory.get_full_steps()

            return RunResult(
                output=output,
                token_usage=token_usage,
                messages=messages,
                timing=Timing(start_time=run_start_time, end_time=time.time()),
                state=state,
            )

        return output


class PZBaseManagedAgent(PZBaseAgent):

    def __call__(self, task: str, **kwargs):
        """Adds additional prompting for the managed agent, runs it, and wraps the output.
        This method is called only by a managed agent.
        """
        full_task = populate_template(
            self.prompt_templates["managed_agent"]["task"],
            variables=dict(name=self.name, task=task, context_description=self.context_description),
        )
        result = self.run(full_task, **kwargs)
        report = result.output if isinstance(result, RunResult) else result
        answer = populate_template(
            self.prompt_templates["managed_agent"]["report"], variables=dict(name=self.name, final_answer=report)
        )
        if self.provide_run_summary:
            answer += "\n\nFor more detail, find below a summary of this agent's work:\n<summary_of_work>\n"
            for message in self.write_memory_to_messages(summary_mode=True):
                content = message.content
                answer += "\n" + truncate_content(str(content)) + "\n---"
            answer += "\n</summary_of_work>"
        return answer


class DataDiscoveryAgent(PZBaseManagedAgent):
    def __init__(self, run_id: str, context_description: str, *args, **kwargs):
        self.description = """A team member that will search a data repository to find files which help to answer your question.
    Ask him for all your questions that require searching a repository of relevant data.
    Provide him as much context as possible, in particular if you need to search on a specific timeframe!
    And don't hesitate to provide him with a complex search task, like finding a difference between two files.
    Your request must be a real sentence, not a keyword search! Like "Find me this information (...)" rather than a few keywords.
        """
        prompt_templates = PromptTemplates(
            system_prompt=CODE_AGENT_SYSTEM_PROMPT,
            planning=PlanningPromptTemplate(
                initial_plan=DATA_DISCOVERY_AGENT_INITIAL_PLAN_PROMPT,
                update_plan_pre_messages=DATA_DISCOVERY_AGENT_UPDATE_PLAN_PRE_MESSAGES_PROMPT,
                update_plan_post_messages=DATA_DISCOVERY_AGENT_UPDATE_PLAN_POST_MESSAGES_PROMPT,
            ),
            managed_agent=ManagedAgentPromptTemplate(task=DATA_DISCOVERY_AGENT_TASK_PROMPT, report=DATA_DISCOVERY_AGENT_REPORT_PROMPT),
            final_answer=FinalAnswerPromptTemplate(pre_messages=FINAL_ANSWER_PRE_MESSAGES_PROMPT, post_messages=FINAL_ANSWER_POST_MESSAGES_PROMPT),
        )

        super().__init__(
            *args,
            run_id=run_id,
            context_description=context_description,
            prompt_templates=prompt_templates,
            max_steps=20,
            verbosity_level=2,
            planning_interval=4,
            name="data_discovery_agent",
            description=self.description,
            provide_run_summary=True,
            **kwargs,
        )
        self.prompt_templates["managed_agent"]["task"] += """Additionally, if after some searching you find out that you need more information to answer the question, you can use `final_answer` with your request for clarification as argument to request for more information."""


class SearchManagerAgent(PZBaseAgent):
    def __init__(self, run_id: str, context_description: str, *args, **kwargs):
        prompt_templates = PromptTemplates(
            system_prompt=CODE_AGENT_SYSTEM_PROMPT,
            planning=PlanningPromptTemplate(
                initial_plan=DATA_DISCOVERY_AGENT_INITIAL_PLAN_PROMPT,
                update_plan_pre_messages=DATA_DISCOVERY_AGENT_UPDATE_PLAN_PRE_MESSAGES_PROMPT,
                update_plan_post_messages=DATA_DISCOVERY_AGENT_UPDATE_PLAN_POST_MESSAGES_PROMPT,
            ),
            managed_agent=ManagedAgentPromptTemplate(task=DATA_DISCOVERY_AGENT_TASK_PROMPT, report=DATA_DISCOVERY_AGENT_REPORT_PROMPT),
            final_answer=FinalAnswerPromptTemplate(pre_messages=FINAL_ANSWER_PRE_MESSAGES_PROMPT, post_messages=FINAL_ANSWER_POST_MESSAGES_PROMPT),
        )
        super().__init__(
            *args,
            run_id=run_id,
            context_description=context_description,
            prompt_templates=prompt_templates,
            max_steps=12,
            verbosity_level=2,
            additional_authorized_imports=["*"],
            planning_interval=4,
            return_full_result=True,
            **kwargs,
        )

# class ManagerAgent(CodeAgent):

#     def _step_stream(self, memory_step: ActionStep) -> Generator[ChatMessageStreamDelta | ActionOutput | ToolOutput]:
#         """
#         Perform one step in the ReAct framework: the agent thinks, acts, and observes the result.
#         Yields ChatMessageStreamDelta during the run if streaming is enabled.
#         At the end, yields either None if the step is not final, or the final answer.
#         """
#         raise NotImplementedError("This method should be implemented in child classes")


from __future__ import annotations

from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.models import OperatorCostEstimates, RecordOpStats
from palimpzest.query.operators.physical import PhysicalOperator


class LimitScanOp(PhysicalOperator):
    def __init__(self, limit: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.limit = limit

    def __str__(self):
        op = super().__str__()
        op += f"    Limit: {self.limit}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"limit": self.limit, **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"limit": self.limit, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the limit takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=min(self.limit, source_op_cost_estimates.cardinality),
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidate: DataRecord) -> DataRecordSet:
        # NOTE: execution layer ensures that no more than self.limit
        #       records are returned to the user by this operator.
        # create new DataRecord
        dr = DataRecord.from_parent(schema=candidate.schema, data_item={}, parent_record=candidate)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=0.0,
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])




CONTEXT_SEARCH_PROMPT = """You are a helpful agent whose job is to propose search queries that will assist in finding information that is relevant to performing a computation.

Please propose a concise sentence which will be used to search for relevant information for the following computation instruction.

INSTRUCTION: {instruction}

SEARCH QUERY: 
"""



import logging
import os
from enum import Enum

from dotenv import load_dotenv

from palimpzest.core.data.dataset import Dataset
from palimpzest.core.elements.records import DataRecordCollection
from palimpzest.query.execution.execution_strategy import ExecutionStrategy, SentinelExecutionStrategy
from palimpzest.query.execution.execution_strategy_type import ExecutionStrategyType, SentinelExecutionStrategyType
from palimpzest.query.optimizer.cost_model import SampleBasedCostModel
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.query.optimizer.optimizer_strategy_type import OptimizationStrategyType
from palimpzest.query.processor.config import QueryProcessorConfig
from palimpzest.query.processor.query_processor import QueryProcessor
from palimpzest.utils.model_helpers import get_models
from palimpzest.validator.validator import Validator

logger = logging.getLogger(__name__)


class QueryProcessorFactory:

    @classmethod
    def _convert_to_enum(cls, enum_type: type[Enum], value: str) -> Enum:
        value = value.upper().replace('-', '_')
        try:
            return enum_type[value]
        except KeyError as e:
            raise ValueError(f"Unsupported {enum_type.__name__}: {value}") from e

    @classmethod
    def _normalize_strategies(cls, config: QueryProcessorConfig):
        """
        Convert the string representation of each strategy into its Enum equivalent and throw
        an exception if the conversion fails.
        """
        strategy_types = {
            "execution_strategy": ExecutionStrategyType,
            "sentinel_execution_strategy": SentinelExecutionStrategyType,
            "optimizer_strategy": OptimizationStrategyType,
        }
        for strategy in ["execution_strategy", "sentinel_execution_strategy", "optimizer_strategy"]:
            strategy_str = getattr(config, strategy)
            strategy_type = strategy_types[strategy]
            strategy_enum = None
            if strategy_str is not None:
                try:
                    strategy_enum = cls._convert_to_enum(strategy_type, strategy_str)
                except ValueError as e:
                    raise ValueError(f"""Unsupported {strategy}: {strategy_str}.
                                        The supported strategies are: {strategy_type.__members__.keys()}""") from e
            setattr(config, strategy, strategy_enum)
            logger.debug(f"Normalized {strategy}: {strategy_enum}")

        return config

    @classmethod
    def _config_validation_and_normalization(cls, config: QueryProcessorConfig, train_dataset: dict[str, Dataset] | None, validator : Validator | None):
        if config.policy is None:
            raise ValueError("Policy is required for optimizer")
        
        # only one of progress or verbose can be set; we will default to progress=True
        if config.progress and config.verbose:
            print("WARNING: Both `progress` and `verbose` are set to True, but only one can be True at a time; defaulting to `progress=True`")
            config.verbose = False

        # if the user provides a training dataset, but no validator, create a default validator
        if train_dataset is not None and validator is None:
            validator = Validator()
            logger.info("No validator provided; using default Validator")

        # boolean flag for whether we're performing optimization or not
        optimization = validator is not None

        # handle "auto" default for sentinel execution strategies
        if config.sentinel_execution_strategy == "auto":
            config.sentinel_execution_strategy = "mab" if optimization else None

        # convert the config values for processing, execution, and optimization strategies to enums
        config = cls._normalize_strategies(config)

        # get available models
        available_models = getattr(config, 'available_models', [])
        if available_models is None or len(available_models) == 0:
            available_models = get_models(use_vertex=config.use_vertex, gemini_credentials_path=config.gemini_credentials_path, api_base=config.api_base)

        # remove any models specified in the config
        remove_models = getattr(config, 'remove_models', [])
        if remove_models is not None and len(remove_models) > 0:
            available_models = [model for model in available_models if model not in remove_models]
            logger.info(f"Removed models from available models based on config: {remove_models}")

        # set the final set of available models in the config
        config.available_models = available_models

        if len(config.available_models) == 0:
            raise ValueError("No available models found.")

        openai_key = os.getenv("OPENAI_API_KEY")
        anthropic_key = os.getenv("ANTHROPIC_API_KEY")
        together_key = os.getenv("TOGETHER_API_KEY")
        gemini_key = os.getenv("GEMINI_API_KEY")
        google_key = os.getenv("GOOGLE_API_KEY")

        for model in config.available_models:
            if model.is_openai_model() and not openai_key:
                raise ValueError("OPENAI_API_KEY must be set to use OpenAI models.")
            if model.is_anthropic_model() and not anthropic_key:
                raise ValueError("ANTHROPIC_API_KEY must be set to use Anthropic models.")
            if model.is_together_model() and not together_key:
                raise ValueError("TOGETHER_API_KEY must be set to use Together models.")
            if model.is_google_ai_studio_model() and not (gemini_key or google_key or config.gemini_credentials_path):
                raise ValueError("GEMINI_API_KEY, GOOGLE_API_KEY, or gemini_credentials path must be set to use Google Gemini models.")
            if model.is_vllm_model() and config.api_base is None:
                raise ValueError("api_base must be set to use vLLM models.")

        return config, validator

    @classmethod
    def _create_optimizer(cls, config: QueryProcessorConfig) -> Optimizer:
        return Optimizer(cost_model=SampleBasedCostModel(), **config.to_dict())

    @classmethod
    def _create_execution_strategy(cls, dataset: Dataset, config: QueryProcessorConfig) -> ExecutionStrategy:
        """
        Creates an execution strategy based on the configuration.
        """
        # for parallel execution, set the batch size if there's a limit in the query
        limit = dataset.get_limit()
        if limit is not None and config.execution_strategy == ExecutionStrategyType.PARALLEL:
            if config.batch_size is None:
                config.batch_size = limit
                logger.info(f"Setting batch size to query limit: {limit}")
            elif config.batch_size > limit:
                config.batch_size = limit
                logger.info(f"Setting batch size to query limit: {limit} since it was larger than the limit")

        # create the execution strategy
        execution_strategy_cls = config.execution_strategy.value
        return execution_strategy_cls(**config.to_dict())

    @classmethod
    def _create_sentinel_execution_strategy(cls, config: QueryProcessorConfig) -> SentinelExecutionStrategy:
        """
        Creates an execution strategy based on the configuration.
        """
        if config.sentinel_execution_strategy is None:
            return None

        sentinel_execution_strategy_cls = config.sentinel_execution_strategy.value
        return sentinel_execution_strategy_cls(**config.to_dict())

    @classmethod
    def create_processor(
        cls,
        dataset: Dataset,
        config: QueryProcessorConfig | None = None,
        train_dataset: dict[str, Dataset] | None = None,
        validator: Validator | None = None,
    ) -> QueryProcessor:
        """
        Creates a QueryProcessor with specified processing and execution strategies.

        Args:
            dataset: The dataset to process
            config: The user-provided QueryProcessorConfig; if it is None, the default config will be used
            kwargs: Additional keyword arguments to pass to the QueryProcessorConfig
        """
        if config is None:
            config = QueryProcessorConfig()

        # make a copy of the config to avoid modifying the original
        config = config.copy()

        # apply any additional keyword arguments to the config and validate its contents
        config, validator = cls._config_validation_and_normalization(config, train_dataset, validator)

        # update the dataset's types if we're not enforcing types
        if not config.enforce_types:
            dataset.relax_types()
            if train_dataset is not None:
                for _, ds in train_dataset.items():
                    ds.relax_types()

        # create the optimizer, execution strateg(ies), and processor
        optimizer = cls._create_optimizer(config)
        config.execution_strategy = cls._create_execution_strategy(dataset, config)
        config.sentinel_execution_strategy = cls._create_sentinel_execution_strategy(config)
        processor = QueryProcessor(dataset, optimizer, train_dataset=train_dataset, validator=validator, **config.to_dict())

        return processor

    @classmethod
    def create_and_run_processor(
        cls,
        dataset: Dataset,
        config: QueryProcessorConfig | None = None,
        train_dataset: dict[str, Dataset] | None = None,
        validator: Validator | None = None,
    ) -> DataRecordCollection:
        load_dotenv(override=True)
        logger.info(f"Creating processor for dataset: {dataset}")
        processor = cls.create_processor(dataset, config, train_dataset, validator)
        logger.info(f"Created processor: {processor}")

        return processor.execute()



import logging

from palimpzest.constants import Cardinality, Model
from palimpzest.core.data.context import Context, TextFileContext
from palimpzest.core.data.dataset import Dataset
from palimpzest.core.data.iter_dataset import (
    AudioFileDataset,
    HTMLFileDataset,
    ImageFileDataset,
    IterDataset,
    MemoryDataset,
    PDFFileDataset,
    TextFileDataset,
    XLSFileDataset,
)
from palimpzest.core.elements.groupbysig import GroupBySig
from palimpzest.core.lib.schemas import AudioBase64, AudioFilepath, ImageBase64, ImageFilepath, ImageURL
from palimpzest.policy import (
    MaxQuality,
    MaxQualityAtFixedCost,
    MaxQualityAtFixedTime,
    MinCost,
    MinCostAtFixedQuality,
    MinTime,
    MinTimeAtFixedQuality,
    PlanCost,
    Policy,
)
from palimpzest.query.processor.config import QueryProcessorConfig
from palimpzest.validator.validator import Validator

# Initialize the root logger
logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    # constants
    "Cardinality",
    "Model",
    # core
    "GroupBySig",
    "Context",
    "TextFileContext",
    "Dataset",
    "IterDataset",
    "AudioFileDataset",
    "MemoryDataset",
    "HTMLFileDataset",
    "ImageFileDataset",
    "PDFFileDataset",
    "TextFileDataset",
    "XLSFileDataset",
    # schemas
    "AudioBase64",
    "AudioFilepath",
    "ImageBase64",
    "ImageFilepath",
    "ImageURL",
    # policy
    "MaxQuality",
    "MaxQualityAtFixedCost",
    "MaxQualityAtFixedTime",
    "MinCost",
    "MinCostAtFixedQuality",
    "MinTime",
    "MinTimeAtFixedQuality",
    "PlanCost",
    "Policy",
    # query
    "QueryProcessorConfig",
    # validator
    "Validator",
]
